searchData={"items":[{"type":"module","title":"Carbonite","doc":"Carbonite makes it easy to create audit trails for tables in a PostgreSQL database and integrate them into an Elixir application.\n\nCarbonite implements the [Change-Data-Capture](https://en.wikipedia.org/wiki/Change_data_capture) pattern on top of a PostgreSQL database. It uses triggers to automatically record all changes applied to a database table in order to guarantee a complete audit trail of the contained data.\n\nIt is centered around the idea that the database transaction is the natural auditing unit of data mutation. Any mutation on a participating table requires the developer first to record the operation metadata within the same database transaction. The metadata record is associated to the table change records by a constraint.\n\nOn top of its database layer, Carbonite provides an API allowing developers to effortlessly retrieve, process, and purge the audit trails within the Elixir application.","ref":"Carbonite.html"},{"type":"module","title":"How it works - Carbonite","doc":"Carbonite keeps a central `changes` table where all mutations of participating tables are recorded. On each such table a trigger is installed (after `INSERT`, `UPDATE`, and `DELETE` statements - `TRUNCATE` is not supported, see below) which calls a procedure stored within the database. This procedure captures the new or updated data within the `changes` table automatically. The procedure fetches its own per-table settings from another table called `triggers`. These settings customize the procedure's behaviour, for instance, they may exclude certain table columns from being captured.\n\nBesides the `changes` table, storing information about individual database statements (or rather, their impact on the data), Carbonite's `transactions` table stores information on database _transactions_. Within your application logic, you can insert a row into this table after you begin a database transaction, and record arbitrary metadata for it.\n\nApart from the metadata, the `transactions` table houses two identifier columns: `id` and `xact_id`. While `id` is an ordinary, autoincrementing integer primary key, the `xact_id` is set to `pg_current_xact_id()`, PostgreSQL's internal transaction identifier. Each row in the `changes` table is associated to a single row in the `transactions` table, referencing *both these identifiers*.\n\n- The `changes.transaction_id` (referencing `transactions.id`) field is filled with the [\"current value\"](https://www.postgresql.org/docs/13/functions-sequence.html) of the related sequence. This is your regular foreign key column, enough to relate a `changes` record with exactly one `transactions` record. Consistency with the `transactions` table is guaranteed by a foreign key constraint.\n- Additionally, the `changes.transaction_xact_id` field is set to `pg_current_xact_id()`. Together these references ensure that, not only relates each `changes` record to a single `transactions` record, but records in these tables have been inserted *within the same database transaction*. Consistency is ensure by a manual lookup in the trigger procedure.\n\nThis leads to the following interesting properties:\n\n- All `changes` created within a database transaction automatically and implicitly belong to the same record in the `transactions` table, even if they're created separately and agnostic of each other in the application logic. This gives the developer a \"natural\" way to group related changes into events (more on events later).\n- The entry in the `transactions` table _must be created before any `changes`_. Attempting to modify an audited table without prior insertion into the `transactions` table will result in an error.\n\n ℹ️ &ensp;Trigger vs. Write-Ahead-Log / Logical Decoding / Extensions \n\nExisting solutions for CDC on top of a PostgreSQL database (e.g. [Debezium](https://debezium.io/)) often tail the [Write-Ahead-Log](https://www.postgresql.org/docs/13/wal-intro.html) or use extensions (e.g. [pgaudit](https://www.pgaudit.org/)) instead of using triggers to create change records. Both approaches have their own advantages and disadvantages, and one should try to understand the differences before making up their mind.\n\n* One of the main selling points of triggers is that they are executed as part of the normal transaction logic in the database, hence they benefit from the usual [atomicity and consistency guarantees](https://en.wikipedia.org/wiki/ACID) of relational databases. For audit triggers, among other things this means that data mutations and their audit trail are committed as an atomic unit.\n* At the same time, this property of triggers strongly couples the auditing logic to the business operation. For instance, if the audit trigger raises an error, the entire transaction is aborted. And vice versa, if the business operation aborts, nothing is audited.\n* Additionally, auditing unavoidably has an impact on the database performance. A common criticism of audit triggers is the performance penalty they incur on otherwise \"simple\" database operations, which is said to be much higher than just tailing the WAL. Carbonite tries to limit the work done in its trigger, but a few indexes and tables are nonetheless touched.\n* Another advantage of triggers is their universality: You should be able to run Carbonite's migrations on any hosted PostgreSQL instance without the need to tweak its configuration or install custom extensions before.\n* The primary advantage of triggers for Carbonite, though, is that we can make use of PostgreSQL's internal transaction id: It allows us to group related changes together and record metadata for the group. This and the simple tooling around it is not (as easily) achievable with a fully decoupled system. However, if you do not need this feature of Carbonite, please be sure to consider other solutions as well.\n\n ℹ️ &ensp;Why  TRUNCATE  is not supported \n\nAs the last of the 4 primary data mutating SQL commands, `TRUNCATE` allows to delete all data from a table or a set of tables. It can be instrumented using triggers on the statement level (`FOR EACH STATEMENT` instead of `FOR EACH ROW`), which means the trigger procedure executes only once for the statement and without any data - in contrast to, for instance, an `UPDATE` statement, which also mutates multiple rows but fires the procedure once for each row. While there might be value in auditing `TRUNCATE` statements, the behaviour of the trigger procedure would be quite different from the other commands. For a rarely used SQL command, we chose against this additional complexity and not to support `TRUNCATE` in Carbonite.","ref":"Carbonite.html#module-how-it-works"},{"type":"module","title":"Installation - Carbonite","doc":"#","ref":"Carbonite.html#module-installation"},{"type":"module","title":"Requirements - Carbonite","doc":"Due to its use of [`pg_current_xact_id`](https://www.postgresql.org/docs/13/functions-info.html#FUNCTIONS-PG-SNAPSHOT), Carbonite requires PostgreSQL *version 13 or above*. If you see an error message like the following, your PostgreSQL installation is too old:\n\n```\n** (Postgrex.Error) ERROR 42704 (undefined_object) type \"xid8\" does not exist\n```\n\n#","ref":"Carbonite.html#module-requirements"},{"type":"module","title":"Hex dependency - Carbonite","doc":"```elixir\n# mix.exs\ndef deps do\n  [\n    {:carbonite, \"~> 0.14.0\"}\n  ]\nend\n```","ref":"Carbonite.html#module-hex-dependency"},{"type":"module","title":"Getting started - Carbonite","doc":"As much of Carbonite's logic lives in database functions and triggers, to get started, we first need to create a migration using Ecto.\n\n#","ref":"Carbonite.html#module-getting-started"},{"type":"module","title":"Creating the initial migration - Carbonite","doc":"Carbonite contains a Mix task that generates the initial migration for you. Please open the generated file and edit it according to your needs.\n\nSee `Carbonite.Migrations` for more information on migrations.\n\n```sh\nmix carbonite.gen.initial_migration -r MyApp.Repo\n```\n\nThe final migration should look something like this:\n\n```elixir\n# priv/repo/migrations/20240404201534_install_carbonite.exs\ndefmodule MyApp.Repo.Migrations.InstallCarbonite do\n  use Ecto.Migration\n\n  def up do\n    Carbonite.Migrations.up(1..9)\n\n    # For each table that you want to capture changes of, you need to install the trigger.\n    Carbonite.Migrations.create_trigger(:rabbits)\n\n    # Optionally you may configure the trigger inside the migration.\n    Carbonite.Migrations.put_trigger_config(:rabbits, :excluded_columns, [\"age\"])\n\n    # Optional outbox to process transactions later.\n    Carbonite.Migrations.create_outbox(\"rabbit_holes\")\n  end\n\n  def down do\n    # Remove outbox again.\n    Carbonite.Migrations.drop_outbox(\"rabbit_holes\")\n\n    # Remove all triggers before dropping the schema.\n    Carbonite.Migrations.drop_trigger(:rabbits)\n\n    # Drop the Carbonite tables.\n    Carbonite.Migrations.down(9..1)\n  end\nend\n```\n\n#","ref":"Carbonite.html#module-creating-the-initial-migration"},{"type":"module","title":"Updates - Carbonite","doc":"When a new Carbonite version is released, it may contain updates to the database schema. As these are announced in the [Changelog](CHANGELOG.md), you each time need to create a migration in your host application like the following.\n\n```elixir\n# priv/repo/migrations/20210704201534_update_carbonite.exs\ndefmodule MyApp.Repo.Migrations.UpdateCarbonite do\n  use Ecto.Migration\n\n  def up do\n    Carbonite.Migrations.up(3)\n  end\n\n  def down do\n    Carbonite.Migrations.down(3)\n  end\nend\n```\n\nApplying multiple Carbonite migrations in a single host migration is fine. Note that for each of your Carbonite \"partitions\" (see below), you need to run each Carbonite migration.\n\n#","ref":"Carbonite.html#module-updates"},{"type":"module","title":"Trigger configuration - Carbonite","doc":"The behaviour of the capture trigger is customizable per table by manipulating the settings in the `triggers` table. Often you will want to update these settings within a migration, as well, which is why Carbonite provides the small helper function `Carbonite.Migrations.put_trigger_config/4` that updates the settings using plain SQL statements.\n\n##","ref":"Carbonite.html#module-trigger-configuration"},{"type":"module","title":"Primary Key Columns - Carbonite","doc":"To speed up version lookups for a specific record, Carbonite copies its primary key(s) to the `table_pk` column of the `changes` table. The table keeps an index on this column together with the table prefix and name.\n\nBy default, Carbonite will try to copy the `:id` column of the source table. If your table does not have a primary key, has a primary key with a different name, or has a composite primary key, you can override this using the `primary_key_columns` option.\n\n```elixir\n# Disable PK copying\nCarbonite.Migrations.put_trigger_config(:rabbits, :primary_key_columns, [])\n\n# Different name\nCarbonite.Migrations.put_trigger_config(:rabbits, :primary_key_columns, [\"identifier\"])\n\n# Composite PK\nCarbonite.Migrations.put_trigger_config(:rabbits, :primary_key_columns, [\"house\", \"apartment_no\"])\n```\n\nSince the `changes` table keeps versions of a multitude of different source tables, primary keys are first cast to string (the `table_pk` column has type `VARCHAR[]`). For composite primary keys, set the `primary_key_columns` option to an array as shown above. Each component of a compound primary key will be cast to string before the components are joined into the array.\n\n##","ref":"Carbonite.html#module-primary-key-columns"},{"type":"module","title":"Excluded and Filtered Columns - Carbonite","doc":"In case your table contains sensitive data or data otherwise undesirable for change capturing, you can exclude columns using the `excluded_columns` option. Excluded columns will not appear in the captured data. If an `UPDATE` on a table solely touches excluded columns, the entire `UPDATE` will not be recorded.\n\n```elixir\nCarbonite.Migrations.put_trigger_config(:rabbits, :excluded_columns, [\"age\"])\n```\n\nIf you still want to capture changes to a column (in the `changed` field), but don't need the exact data, you can make it a \"filtered\" column. These columns appear as `[FILTERED]` in the `data` field.\n\n```elixir\nCarbonite.Migrations.put_trigger_config(:rabbits, :filtered_columns, [\"age\"])\n```\n\n##","ref":"Carbonite.html#module-excluded-and-filtered-columns"},{"type":"module","title":"Persisting replaced data on `UPDATE` statements - Carbonite","doc":"By default, the `changes` table will track the current version of a record in its `data` field and a list of changed fields in the `changed` column. Optionally, you can decide to store the replaced data (the \"diff\") in the `changed_from` field, by enabling the `store_changed_from` option for a trigger. This defaults to `false`.\n\n```elixir\nCarbonite.Migrations.put_trigger_config(:rabbits, :store_changed_from, true)\n```\n\n#","ref":"Carbonite.html#module-persisting-replaced-data-on-update-statements"},{"type":"module","title":"Partitioning the Audit Trail - Carbonite","doc":"Carbonite can install its tables into multiple database schemas using the `carbonite_prefix` option. You can use this feature to \"partition\" your captured data.\n\n```elixir\nCarbonite.Migrations.up(1, carbonite_prefix: \"carbonite_animals\")\nCarbonite.Migrations.create_trigger(:rabbits, carbonite_prefix: \"carbonite_animals\")\n```\n\nBasically all of Carbonite's functions accept the same `carbonite_prefix` option to target a particular partition. If desired, tables can participate in multiple partitions by adding multiple triggers on to them. Keep in mind that each partition will need to be processed and purged separately, resulting in multiple streams of change data in your external storage.","ref":"Carbonite.html#module-partitioning-the-audit-trail"},{"type":"module","title":"Inserting a `Transaction` - Carbonite","doc":"In your application logic, before modifying an audited table like `rabbits`, you need to first create a `Carbonite.Transaction` record.\n\n#","ref":"Carbonite.html#module-inserting-a-transaction"},{"type":"module","title":"With Ecto.Multi - Carbonite","doc":"The easiest way to do so is using `Carbonite.Multi.insert_transaction/2` within an `Ecto.Multi` operation:\n\n```elixir\nEcto.Multi.new()\n|> Carbonite.Multi.insert_transaction(%{meta: %{type: \"rabbit_inserted\"}})\n|> Ecto.Multi.insert(:rabbit, &MyApp.Rabbit.create_changeset(&1.params))\n|> MyApp.Repo.transaction()\n```\n\nAs you can see, the `Carbonite.Transaction` is a great place to store metadata for the operation. A `type` field can be used to categorize the transactions. A `user_id` would be a good candidate for a transaction log, as well.\n\n#","ref":"Carbonite.html#module-with-ecto-multi"},{"type":"module","title":"Without Ecto.Multi - Carbonite","doc":"If you have a situation where you want to insert a transaction without `Ecto.Multi`, you can use `Carbonite.insert_transaction/3`.\n\n``` elixir\nMyApp.Repo.transaction(fn ->\n  {:ok, _} = Carbonite.insert_transaction(MyApp.Repo, %{meta: %{type: \"rabbit_inserted\"}})\n\n  params\n  |> MyApp.Rabbit.create_changeset()\n  |> MyApp.Repo.insert!\nend)\n```\n\n#","ref":"Carbonite.html#module-without-ecto-multi"},{"type":"module","title":"Building a changeset for manual insertion - Carbonite","doc":"Another alternative is creating a changeset for a `Carbonite.Transaction` using `Carbonite.Transaction.changeset/1`:\n\n```elixir\nMyApp.Repo.transaction(fn ->\n  %{meta: %{type: \"rabbit_inserted\"}}\n  |> Carbonite.Transaction.changeset()\n  |> MyApp.Repo.insert!()\n\n  # ...\nend)\n```\n\n#","ref":"Carbonite.html#module-building-a-changeset-for-manual-insertion"},{"type":"module","title":"Setting metadata outside of the transaction - Carbonite","doc":"In case you do not have access to metadata you want to persist in the `Carbonite.Transaction` at the code site where you create it, you can use `Carbonite.Transaction.put_meta/2` to store metadata in the _process dictionary_. This metadata is merged into the metadata given to `Carbonite.Multi.insert_transaction/2`.\n\n```elixir\n# e.g., in a controller or plug\nCarbonite.Transaction.put_meta(:user_id, ...)\n```\n\n#","ref":"Carbonite.html#module-setting-metadata-outside-of-the-transaction"},{"type":"module","title":"Transactions in data migrations - Carbonite","doc":"If you manipulate data inside your transactions, as usual a `Carbonite.Transaction` needs to be inserted before any other statements. You can use `Carbonite.Migrations.insert_migration_transaction/1` to insert a transaction with a `meta` attribute populated from the migration module.\n\n```elixir\nimport Carbonite.Migrations\n\ndef change do\n  insert_migration_transaction()\n\n  execute(\"UPDATE ...\")\nend\n```\n\n### \"Empty\" transactions when no changes have been recorded\n\nBy default, Carbonite will store the `Carbonite.Transaction` record regardless of whether in the corresponding database transaction any changes were recorded or not. In fact, there is nothing special about the behaviour of the following code:\n\n```elixir\nEcto.Multi.new()\n|> Carbonite.Multi.insert_transaction(%{meta: %{type: \"rabbit_inserted\"}})\n|> Ecto.Multi.run(:rabbit, &maybe_insert_rabbit/2)\n|> MyApp.Repo.transaction()\n```\n\nDepending on the behaviour of `maybe_insert_rabbit/2`, the transaction may result in one of these outcomes:\n\n1. The transaction succeeds and the rabbit is inserted. A `Carbonite.Transaction` is recorded alongside a `Carbonite.Change`.\n2. The transaction succeeds but *no rabbit is inserted*. The `Carbonite.Transaction` is recorded but will be empty, that is, not associated to any `Carbonite.Change` records.\n3. The transaction fails and is rolled back. The `Carbonite.Transaction` is not persisted.\n\nOutcome (1) and (3) above are engrained in Carbonite's trigger logic and can not be changed.\n\nThe second outcome can be considered \"intended behaviour\" and there are good reasons for keeping the `Carbonite.Transaction` around. However, if your use-case indicates that you should not store `Carbonite.Transactions` when nothing was changed, you need to:\n\n1. Set the trigger to `INITIALLY DEFERRED` when it is created.\n2. Re-order your application logic.\n3. Only insert the transaction when needed.\n\n```elixir\n# In the migration where the trigger is created\nCarbonite.Migrations.create_trigger(:rabbits, initially: :deferred)\n```\n\n```elixir\nEcto.Multi.new()\n|> Ecto.Multi.run(:rabbit, &maybe_insert_rabbit/2)\n|> Ecto.Multi.run(:carbonite_transaction, &maybe_insert_carbonite_transaction/2)\n|> MyApp.Repo.transaction()\n```\n\nSee `Carbonite.Migrations.create_trigger/2` for further information.","ref":"Carbonite.html#module-transactions-in-data-migrations"},{"type":"module","title":"Retrieving data - Carbonite","doc":"Of course, persisting the audit trail is not an end in itself. At some point you will want to read the data back and make it accessible to the user. `Carbonite.Query` offers a small suite of helper functions that make it easier to query the database for `Transaction` and `Change` records.\n\n#","ref":"Carbonite.html#module-retrieving-data"},{"type":"module","title":"Fetching transactions - Carbonite","doc":"The `Carbonite.Query.transactions/1` function constructs an `Ecto.Query` for loading `t:Carbonite.Transaction.t/0` records from the database, optionally preloading their included changes. The query can be further refined to limit the result set.\n\n```elixir\nCarbonite.Query.transactions()\n|> Ecto.Query.where([t], t.inserted_at > ^earliest)\n|> MyApp.Repo.all()\n```\n\n#","ref":"Carbonite.html#module-fetching-transactions"},{"type":"module","title":"Fetching changes of individual records - Carbonite","doc":"The `Carbonite.Query.changes/2` function constructs an `t:Ecto.Query.t/0` from a schema struct, loading all changes stored for the given source record.\n\n```elixir\n%MyApp.Rabbit{id: 1}\n|> Carbonite.Query.changes()\n|> MyApp.Repo.all()\n```","ref":"Carbonite.html#module-fetching-changes-of-individual-records"},{"type":"module","title":"Processing data - Carbonite","doc":"Storing event series in a relational database comes with the usual scaling issues, so at some point you may find it necessary to offload the captured data to an external data store. For processing, exporting, and later purging the captured transactions, Carbonite includes what is called an \"Event Outbox\", or in fact as many of those as you need.\n\nEssentially, a `Carbonite.Outbox` is a named cursor on the ordered list of transactions stored in an audit trail. Each outbox advances by calling a user-supplied processor function on a batch of transactions. The batch can be filtered and limited in size, but is always ordered ascending by the transaction's `id` attribute, so transactions are processed *roughly* in order of their insertion (see below for a discussion on caveats of this solution). When a batch is processed, the outbox remembers its last position - the last transaction that has been processed - so the following call to the outbox can continue where the last has left of. Besides, the outbox persists a user-definable \"memo\", an arbitrary map that allows to feed output of the previous processor run into the next one.\n\n#","ref":"Carbonite.html#module-processing-data"},{"type":"module","title":"Creating an Outbox - Carbonite","doc":"Create an outbox within a migration:\n\n```elixir\nCarbonite.Migrations.create_outbox(\"rabbit_holes\")\n```\n\nYou can create more than one outbox if you need multiple processors, e.g. for multiple target systems for your data.\n\n#","ref":"Carbonite.html#module-creating-an-outbox"},{"type":"module","title":"Processing - Carbonite","doc":"You can process an outbox by calling `Carbonite.process/4` with the outbox name and a processor callback function.\n\n```elixir\nCarbonite.process(MyApp.Repo, \"rabbit_holes\", fn transactions, _memo ->\n  for transaction  \n      for transaction  ⚠️&ensp;Transactionality & mutual exclusion \n\nAs the outbox processor is likely to call external services as part of its work, `Carbonite.process/4` does not begin a transaction itself, and consequently does not acquire a lock on the outbox record. In other words, users have to ensure that they don't accidentally run `Carbonite.process/4` for the same outbox concurrently, for instance making use of uniqueness options of the used job processor (e.g. [`unique` in Oban](https://hexdocs.pm/oban/Oban.html#module-unique-jobs)).\n\nThe absence of transactionality also means that any exception raised within the processor function will immediately abort the `Carbonite.process/4` call without writing the current outbox position to disk. As a result, transactions may be processed again in the next run. Please make sure your receiving system or database can handle duplicate messages.\n\n ⚠️&ensp;Long running / parallel transactions and the outbox order \n\nWhen a `Carbonite.Transaction` record is created at the beginning of an operation in your application, it records the \"current\" sequence value in its `id` field. The record in the `transactions` table becomes visible to other transactions when the current transaction is committed.\n\nA few observations can be made from this:\n\n* When ordered by their `id` field, the transactions will be roughly sorted by the time the corresponding operation was started, not when it was committed.\n* Two transactions running in parallel may be committed \"out of order\", i.e. one with the larger `id` may be committed before the smaller `id` transaction if that has a longer runtime.\n\nThe latter point is crucial: For instance, if two transactions with `id=1` and `id=2` run in parallel and `id=2` finishes before `id=1`, an outside viewer can already see `id=2` in the database before `id=1` is committed. If this outside viewer happens to be an outbox processing job, transaction with `id=1` might be skipped and never looked at again. To mitigate this issue, `Carbonite.process/4` has a `min_age` option which excludes transaction younger than a certain from the processing batch (5 minutes by default, increase this is you expect longer running transactions).\n\n#","ref":"Carbonite.html#module-processing"},{"type":"module","title":"Purging - Carbonite","doc":"Now that you have successfully exported your audit data, you may delete old transactions from the primary database. `Carbonite.purge/2` deletes records that have been processed by all existing outboxes.\n\n```elixir\n# Deletes records whose id field is less than the transaction_id on each outbox.\nCarbonite.purge(MyApp.Repo)\n```","ref":"Carbonite.html#module-purging"},{"type":"module","title":"Testing / Bypassing Carbonite - Carbonite","doc":"One of Carbonite's key features is that it is virtually impossible to forget to record a change to a table (due to the trigger) or to forget to insert an enclosing `Carbonite.Transaction` beforehand (due to the foreign key constraint between `changes` and `transactions`). However, in some circumstances it may be desirable to temporarily switch off change capturing. One such situation is the use of factories (e.g. ExMachina) inside your test suite: Inserting a transaction before each factory call quickly becomes cumbersome and will unnecessarily increase execution time. Additionally, the use of Ecto's SQL sandbox means that your factory calls are contained in the same transaction as your domain logic, which means you cannot simply insert a `Carbonite.Transaction` before your factories if you intend to assert on the one inserted by your domain logic.\n\nTo bypass the capture trigger, Carbonite's trigger configuration provides a\ntoggle mechanism consisting of two factors: the `triggers.mode` field defining\nthe default mode and the `Carbonite.override_mode/2` function. The former you\nset while installing the trigger on a table in a migration, while the latter\nallows to \"override\" whatever has been set, at runtime and limited to the current\ntransaction. If you are using Ecto's SQL sandbox for running transactional\ntests, this means the override is going to be active until the end of the test\ncase.\n\nAs a result, you have two options:\n\n1. Leave the `mode` at the default value of `:capture` and *turn off* capturing as needed by switching to \"override mode\". This means for every test case / test setup block where you wish to bypass change capturing, you explicitly disable the trigger before any database calls. This approach has the benefit that you still capture all changes by default, and can't miss to test a code path that (in production) would require a `Carbonite.Transaction`. It is, however, pretty expensive at ~1 additional SQL call per test case.\n2. Set the `mode` to `:ignore` on all triggers in your `:test` environment and instead selectively *turn on* capturing in test cases where you want to assert on the captured data. For instance, you can set the trigger mode in your migration based on the Mix environment. This approach is cheaper as it does not require any action in your tests by default. Yet you should make sure that you test all code paths that do mutate change-captured tables, in order to assert that each of these inserts a transaction as well.\n\n#","ref":"Carbonite.html#module-testing-bypassing-carbonite"},{"type":"module","title":"Option 1: Bypassing change capture selectively - Carbonite","doc":"As an example for option 1, consider the following:\n\n```elixir\n# test/support/carbonite_helpers.ex\ndefmodule MyApp.CarboniteHelpers do\n  def with_ignored_carbonite(fun) do\n    Carbonite.override_mode(MyApp.Repo, to: :ignore)\n    result = fun.()\n    Carbonite.override_mode(MyApp.Repo, to: :capture)\n    result\n  end\nend\n\n# test/some_test.exs\ntest \"my_operation/0\" do\n  rabbit = with_ignored_carbonite(fn -> insert_rabbit() end)\n\n  my_operation(rabbit)\n\n  # You can assert on the Carbonite.Transaction, e.g. like this:\n  meta =\n    Carbonite.Query.current_transaction()\n    |> MyApp.Repo.one!()\n    |> Map.fetch!(:meta)\n\n  assert meta = %{\"type\" => \"some_operation\"}\nend\n\ntest \"my_other_operation/0\" do\n  rabbit = insert_rabbit()\n\n  # this will raise if you forgot to insert a transaction in my_other_operation/0\n  my_operation()\nend\n```\n\n#","ref":"Carbonite.html#module-option-1-bypassing-change-capture-selectively"},{"type":"module","title":"Option 2: Bypassing change capture by default - Carbonite","doc":"The following code snippet illustrates the second approach:\n\n```elixir\n# config/config.exs\nconfig :my_app, carbonite_mode: :capture\n\n# config/test.exs\nconfig :my_app, carbonite_mode: :ignore\n\n# priv/repo/migrations/000000000000_install_carbonite.exs\ndefmodule MyApp.Repo.Migrations.InstallCarbonite do\n  @mode Application.compile_env!(:my_app, :carbonite_mode)\n\n  def up do\n    # ...\n    Carbonite.Migrations.create_trigger(:rabbits)\n    Carbonite.Migrations.put_trigger_config(:rabbits, :mode, @mode)\n  end\nend\n\n# test/support/carbonite_helpers.ex\ndefmodule MyApp.CarboniteHelpers do\n  def enable_carbonite do\n    Carbonite.override_mode(MyApp.Repo, to: :capture)\n  end\nend\n\n# test/some_test.exs\ntest \"my_operation/0\" do\n  rabbit = insert_rabbit()\n\n  enable_carbonite()\n\n  my_operation()\n\n  # Assert on the Carbonite.Transaction...\nend\n\ntest \"my_other_operation/0\" do\n  rabbit = insert_rabbit()\n\n  # this will pass regardless of whether you inserted a transaction in my_other_operation/0 (!)\n  my_operation()\nend\n```","ref":"Carbonite.html#module-option-2-bypassing-change-capture-by-default"},{"type":"function","title":"Carbonite.default_prefix/0","doc":"Returns the default audit trail prefix.","ref":"Carbonite.html#default_prefix/0"},{"type":"function","title":"Carbonite.fetch_changes/2","doc":"Fetches all changes of the current transaction from the database.\n\nMake sure to run this within a transaction.","ref":"Carbonite.html#fetch_changes/2"},{"type":"function","title":"Parameters - Carbonite.fetch_changes/2","doc":"* `repo` - the Ecto repository\n* `opts` - optional keyword list","ref":"Carbonite.html#fetch_changes/2-parameters"},{"type":"function","title":"Options - Carbonite.fetch_changes/2","doc":"* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.html#fetch_changes/2-options"},{"type":"function","title":"Carbonite.insert_transaction/3","doc":"Inserts a `t:Carbonite.Transaction.t/0` into the database.\n\nMake sure to run this within a transaction.","ref":"Carbonite.html#insert_transaction/3"},{"type":"function","title":"Parameters - Carbonite.insert_transaction/3","doc":"* `repo` - the Ecto repository\n* `params` - map of params for the `Carbonite.Transaction` (e.g., `:meta`)\n* `opts` - optional keyword list","ref":"Carbonite.html#insert_transaction/3-parameters"},{"type":"function","title":"Options - Carbonite.insert_transaction/3","doc":"* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.html#insert_transaction/3-options"},{"type":"function","title":"Multiple inserts in the same transaction - Carbonite.insert_transaction/3","doc":"Normally, you should have exactly one `insert_transaction/3` call per database transaction. In\npractise, there are two scenarios in this function may be called multiple times:\n\n1. If an operation A, which calls `insert_transaction/3`, sometimes is nested within an outer\n   operation B, which also calls `insert_transaction/3`.\n2. In tests using Ecto's SQL sandbox, subsequent calls to transactional operations (even to the\n   same operation twice) are wrapped inside the overarching test transaction, and hence also\n   effectively call `insert_transaction/3` within the same transaction.\n\nWhile the first scenario can be resolved using appropriate control flow (e.g. by conditionally\ndisabling the inner `insert_transaction/3` call), the second scenario is quite common and often\nunavoidable.\n\nTherefore, `insert_transaction/3` **ignores** subsequent calls within the same database\ntransaction (equivalent to `ON CONFLICT DO NOTHING`), **discarding metadata** passed to all\ncalls but the first.","ref":"Carbonite.html#insert_transaction/3-multiple-inserts-in-the-same-transaction"},{"type":"function","title":"Carbonite.override_mode/2","doc":"Sets the current transaction to \"override mode\" for all tables in the audit log.","ref":"Carbonite.html#override_mode/2"},{"type":"function","title":"Parameters - Carbonite.override_mode/2","doc":"* `repo` - the Ecto repository\n* `opts` - optional keyword list","ref":"Carbonite.html#override_mode/2-parameters"},{"type":"function","title":"Options - Carbonite.override_mode/2","doc":"* `to` - allows to specify the target mode\n* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.html#override_mode/2-options"},{"type":"function","title":"Carbonite.process/4","doc":"Processes an outbox queue.\n\nThis function sends chunks of persisted transactions to a user-supplied processing function. It\nlooks up the current \"reading position\" from a given `Carbonite.Outbox` and yields transactions\nmatching the given filter criteria (`min_age`, etc.) until either the input source is exhausted\nor a processing function application returns `:halt`.\n\nReturns either `{:ok, outbox}` or `{:halt, outbox}` depending on whether processing was halted\nexplicitly or due the exhausted input source.\n\nSee `Carbonite.Query.outbox_queue/2` for query options.","ref":"Carbonite.html#process/4"},{"type":"function","title":"Examples - Carbonite.process/4","doc":"Carbonite.process(MyApp.Repo, \"rabbit_holes\", fn [transaction], _memo ->\n      # The transaction has its changes preloaded.\n      transaction\n      |> MyApp.Foo.serialize()\n      |> MyApp.Foo.send_to_external_database()\n\n      :cont\n    end)\n\n#","ref":"Carbonite.html#process/4-examples"},{"type":"function","title":"Memo passing - Carbonite.process/4","doc":"The `memo` is useful to carry data between each processor application. Let's say you wanted to\ngenerate a hashsum chain on your processed data:\n\n    Carbonite.process(MyApp.Repo, \"rabbit_holes\", fn [transaction], %{\"checksum\" => checksum} ->\n      {payload, checksum} = MyApp.Foo.serialize_and_hash(transaction, checksum)\n\n      MyApp.Foo.send_to_external_database(payload)\n\n      {:cont, memo: %{\"checksum\" => checksum}}\n    end)\n\n#","ref":"Carbonite.html#process/4-memo-passing"},{"type":"function","title":"Chunking / Limiting - Carbonite.process/4","doc":"The examples above received a single-element list as their first parameter: This is because the\ntransactions are actually processed in \"chunks\" and the default chunk size is 1. If you would\nlike to process more transactions in one chunk, set the `chunk` option:\n\n    Carbonite.process(MyApp.Repo, \"rabbit_holes\", [chunk: 50], fn transactions, _memo ->\n      for transaction   MyApp.Foo.serialize()\n        |> MyApp.Foo.send_to_external_database()\n      end\n\n      :cont\n    end)\n\nThe query that is executed to fetch the data from the database is controlled with the `limit`\noption and is independent of the chunk size.\n\n#","ref":"Carbonite.html#process/4-chunking-limiting"},{"type":"function","title":"Error handling - Carbonite.process/4","doc":"In case you run into an error midway into processing a batch, you may choose to halt processing\nwhile remembering about the last processed transaction. This is equivalent to raising an\nexception from the processing function.\n\n    Carbonite.process(MyApp.Repo, \"rabbit_holes\", fn [transaction], _memo ->\n      case send_to_external_database(transaction) do\n        :ok ->\n          :cont\n\n        {:error, _term} ->\n          :halt\n      end\n    end\n\nYou can, however, if you know the first half of a batch has been processed, still update the\n`memo` and `last_transaction_id`.\n\n    Carbonite.process(MyApp.Repo, \"rabbit_holes\", fn transactions, _memo ->\n      case process_transactions(transactions) do\n        {:error, last_successful_transaction} ->\n          {:halt, last_transaction_id: last_successful_transaction.id}\n\n        :ok ->\n          :cont\n      end\n    end","ref":"Carbonite.html#process/4-error-handling"},{"type":"function","title":"Parameters - Carbonite.process/4","doc":"* `repo` - the Ecto repository\n* `outbox_name` - name of the outbox to process\n* `opts` - optional keyword list\n* `process_func` - see `t:process_func/0` for details","ref":"Carbonite.html#process/4-parameters"},{"type":"function","title":"Options - Carbonite.process/4","doc":"* `min_age` - the minimum age of a record, defaults to 300 seconds (set nil to disable)\n* `limit` - limits the query in size, defaults to 100 (set nil to disable)\n* `filter` - function for refining the batch query, defaults to nil\n* `chunk` - defines the size of the chunk passed to the process function, defaults to 1\n* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.html#process/4-options"},{"type":"function","title":"Carbonite.purge/2","doc":"Deletes transactions that have been fully processed.\n\nSee `Carbonite.Query.outbox_done/1` for query options.\n\nReturns the number of deleted transactions.","ref":"Carbonite.html#purge/2"},{"type":"function","title":"Parameters - Carbonite.purge/2","doc":"* `repo` - the Ecto repository\n* `opts` - optional keyword list","ref":"Carbonite.html#purge/2-parameters"},{"type":"function","title":"Options - Carbonite.purge/2","doc":"* `min_age` - the minimum age of a record, defaults to 300 seconds (set nil to disable)\n* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.html#purge/2-options"},{"type":"type","title":"Carbonite.prefix/0","doc":"","ref":"Carbonite.html#t:prefix/0"},{"type":"type","title":"Carbonite.prefix_option/0","doc":"","ref":"Carbonite.html#t:prefix_option/0"},{"type":"type","title":"Carbonite.process_func/0","doc":"This type defines the callback function signature for `Carbonite.process/3`.\n\nThe processor function receives the current chunk of transactions and the memo of the last\nfunction application, and must return one of\n\n* `:cont` - continue processing\n* `:halt` - stop processing after this chunk\n* `{:cont | :halt, opts}` - cont/halt and set some options\n\nAfter the process function invocation the Outbox is updated with new attributes.","ref":"Carbonite.html#t:process_func/0"},{"type":"type","title":"Options - Carbonite.process_func/0","doc":"Returned options can be:\n\n* `memo` - memo to store on Outbox, defaults to previous memo\n* `last_transaction_id` - last transaction id to remember as processed, defaults to previous\n                          `last_transaction_id` on `:halt`, defaults to last id in current\n                          chunk when `:cont` is returned","ref":"Carbonite.html#t:process_func/0-options"},{"type":"type","title":"Carbonite.process_func_option/0","doc":"","ref":"Carbonite.html#t:process_func_option/0"},{"type":"type","title":"Carbonite.process_option/0","doc":"","ref":"Carbonite.html#t:process_option/0"},{"type":"type","title":"Carbonite.purge_option/0","doc":"","ref":"Carbonite.html#t:purge_option/0"},{"type":"type","title":"Carbonite.repo/0","doc":"","ref":"Carbonite.html#t:repo/0"},{"type":"module","title":"Carbonite.Change","doc":"A `Carbonite.Change` records a mutation on a database table.\n\n`INSERT` statements lead to a `Change` where the `data` field contains the inserted row as a\nJSON object while the `changed` field is an empty list.\n\n`UPDATE` statements contain the updated record in `data` while the `changed` field is a list\nof attributes that have changed. The `changed_from` field may additionally contain a partial\nrepresentation of the replaced record, showing only the diff to the updated record. This\nbehaviour is optional and can be enabled with the `store_changed_from` trigger option.\n\n`DELETE` statements have the deleted data in `data` while `changed` is again an empty list.","ref":"Carbonite.Change.html"},{"type":"type","title":"Carbonite.Change.id/0","doc":"","ref":"Carbonite.Change.html#t:id/0"},{"type":"type","title":"Carbonite.Change.t/0","doc":"","ref":"Carbonite.Change.html#t:t/0"},{"type":"module","title":"Carbonite.Migrations","doc":"Functions to setup Carbonite audit trails in your migrations.","ref":"Carbonite.Migrations.html"},{"type":"function","title":"Carbonite.Migrations.create_outbox/2","doc":"Inserts an outbox record into the database.","ref":"Carbonite.Migrations.html#create_outbox/2"},{"type":"function","title":"Options - Carbonite.Migrations.create_outbox/2","doc":"* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`","ref":"Carbonite.Migrations.html#create_outbox/2-options"},{"type":"function","title":"Carbonite.Migrations.create_trigger/2","doc":"Installs a change capture trigger on a table.","ref":"Carbonite.Migrations.html#create_trigger/2"},{"type":"function","title":"Options - Carbonite.Migrations.create_trigger/2","doc":"* `table_prefix` is the name of the schema the table lives in\n* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`\n* `initially` can be either of `:immediate` or `:deferred`, defaults to `:immediate`","ref":"Carbonite.Migrations.html#create_trigger/2-options"},{"type":"function","title":"Deferred triggers - Carbonite.Migrations.create_trigger/2","doc":"You may create the trigger with `initially: :deferred` option to make it run at the end of your\ndatabase transactions. In combination with application logic that only conditionally inserts\nthe `Carbonite.Transaction`, this can be used to avoid storing \"empty\" transactions, i.e.\n`Carbonite.Transaction` records without associated `Carbonite.Change`.\n\nPlease be aware that this is not recommended by the Carbonite team. Instead we recommend to\nretain the empty transactions and optionally remove them in a preprocessing step or your\noutbox mechanism. Make sure you know how deferred triggers work before using this option.","ref":"Carbonite.Migrations.html#create_trigger/2-deferred-triggers"},{"type":"function","title":"Carbonite.Migrations.down/2","doc":"Rollback a migration.","ref":"Carbonite.Migrations.html#down/2"},{"type":"function","title":"Options - Carbonite.Migrations.down/2","doc":"* `carbonite_prefix` defines the audit trail's schema, defaults to `\"carbonite_default\"`\n* `drop_schema` controls whether the initial migration deletes the schema during rollback","ref":"Carbonite.Migrations.html#down/2-options"},{"type":"function","title":"Carbonite.Migrations.drop_outbox/2","doc":"Removes an outbox record.","ref":"Carbonite.Migrations.html#drop_outbox/2"},{"type":"function","title":"Options - Carbonite.Migrations.drop_outbox/2","doc":"* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`","ref":"Carbonite.Migrations.html#drop_outbox/2-options"},{"type":"function","title":"Carbonite.Migrations.drop_trigger/2","doc":"Removes a change capture trigger from a table.","ref":"Carbonite.Migrations.html#drop_trigger/2"},{"type":"function","title":"Options - Carbonite.Migrations.drop_trigger/2","doc":"* `table_prefix` is the name of the schema the table lives in\n* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`","ref":"Carbonite.Migrations.html#drop_trigger/2-options"},{"type":"function","title":"Carbonite.Migrations.insert_migration_transaction/2","doc":"Inserts a transaction for a data migration.\n\nThe transaction's `meta` attribute is populated with\n\n    {\"type\": \"migration\", direction: \"up\"}\n\n... and additionally the `name` from the parameters.","ref":"Carbonite.Migrations.html#insert_migration_transaction/2"},{"type":"function","title":"Example - Carbonite.Migrations.insert_migration_transaction/2","doc":"defmodule MyApp.Repo.Migrations.SomeDataMigration do\n      use Ecto.Migration\n\n      import Carbonite.Migrations\n\n      def change do\n        insert_migration_transaction()\n\n        execute(\"UPDATE ...\", \"...\")\n      end\n    end\n\nThis works the same for `up/0`/`down/0`-style migrations:\n\n    defmodule MyApp.Repo.Migrations.SomeDataMigration do\n      use Ecto.Migration\n\n      import Carbonite.Migrations\n\n      def up do\n        insert_migration_transaction(\"some-data-migrated\")\n\n        execute(\"INSERT ...\")\n      end\n\n      def down do\n        insert_migration_transaction(\"some-data-rolled-back\")\n\n        execute(\"DELETE ...\")\n      end\n    end","ref":"Carbonite.Migrations.html#insert_migration_transaction/2-example"},{"type":"function","title":"Options - Carbonite.Migrations.insert_migration_transaction/2","doc":"* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`\n* `meta` is a map with additional meta information to store on the transaction","ref":"Carbonite.Migrations.html#insert_migration_transaction/2-options"},{"type":"macro","title":"Carbonite.Migrations.insert_migration_transaction_after_begin/1","doc":"Defines a `c:Ecto.Migration.after_begin/0` implementation for a data migration.\n\nSee `insert_migration_transaction/1` for options.\n\nThis determines the `name` of the transaction from the migration's module name.\n\n     {\"direction\": \"up\",\"name\": \"my_app/repo/migrations/example\", \"type\": \"migration\"}","ref":"Carbonite.Migrations.html#insert_migration_transaction_after_begin/1"},{"type":"macro","title":"Example - Carbonite.Migrations.insert_migration_transaction_after_begin/1","doc":"defmodule MyApp.Repo.Migrations.SomeDataMigration do\n      use Ecto.Migration\n\n      import Carbonite.Migrations\n      insert_migration_transaction_after_begin()\n\n      def change do\n        execute(\"UPDATE ...\")\n      end\n    end\n\nAlternatively, if you define your own migration template module:\n\n    defmodule MyApp.Migration do\n      defmacro __using__ do\n        quote do\n          use Ecto.Migration\n\n          import Carbonite.Migrations\n          insert_migration_transaction_after_begin()\n        end\n      end\n    end","ref":"Carbonite.Migrations.html#insert_migration_transaction_after_begin/1-example"},{"type":"function","title":"Carbonite.Migrations.put_trigger_config/4","doc":"Allows to update a trigger configuration option for a given table.\n\nThis function builds an SQL UPDATE statement that can be used within a database migration to\nupdate a setting stored in Carbonite's `triggers` table without using the `Carbonite.Trigger`\nschema or other application-level code that is prone to change over time. This helps to ensure\nthat your data migrations continue to function, regardless of future updates to Carbonite.","ref":"Carbonite.Migrations.html#put_trigger_config/4"},{"type":"function","title":"Configuration values - Carbonite.Migrations.put_trigger_config/4","doc":"* `primary_key_columns` is a list of columns that form the primary key of the table\n                        (defaults to `[\"id\"]`, set to `[]` to disable)\n* `excluded_columns` is a list of columns to exclude from change captures\n* `filtered_columns` is a list of columns that appear as '[FILTERED]' in the data\n* `store_changed_from` is a boolean defining whether the `changed_from` field should be filled\n* `mode` is either `:capture` or `:ignore` and defines the default behaviour of the trigger","ref":"Carbonite.Migrations.html#put_trigger_config/4-configuration-values"},{"type":"function","title":"Example - Carbonite.Migrations.put_trigger_config/4","doc":"Carbonite.Migrations.put_trigger_config(\"rabbits\", :excluded_columns, [\"name\"])","ref":"Carbonite.Migrations.html#put_trigger_config/4-example"},{"type":"function","title":"Options - Carbonite.Migrations.put_trigger_config/4","doc":"* `table_prefix` is the name of the schema the table lives in\n* `carbonite_prefix` is the schema of the audit trail, defaults to `\"carbonite_default\"`","ref":"Carbonite.Migrations.html#put_trigger_config/4-options"},{"type":"function","title":"Carbonite.Migrations.up/2","doc":"Runs one of Carbonite's migrations.","ref":"Carbonite.Migrations.html#up/2"},{"type":"function","title":"Migration patchlevels - Carbonite.Migrations.up/2","doc":"Make sure that you run all migrations in your host application.\n\n* Initial patch: 1\n* Current patch: 10","ref":"Carbonite.Migrations.html#up/2-migration-patchlevels"},{"type":"function","title":"Options - Carbonite.Migrations.up/2","doc":"* `carbonite_prefix` defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.Migrations.html#up/2-options"},{"type":"type","title":"Carbonite.Migrations.create_trigger_option/0","doc":"","ref":"Carbonite.Migrations.html#t:create_trigger_option/0"},{"type":"type","title":"Carbonite.Migrations.down_option/0","doc":"","ref":"Carbonite.Migrations.html#t:down_option/0"},{"type":"type","title":"Carbonite.Migrations.insert_migration_transaction_option/0","doc":"","ref":"Carbonite.Migrations.html#t:insert_migration_transaction_option/0"},{"type":"type","title":"Carbonite.Migrations.outbox_name/0","doc":"","ref":"Carbonite.Migrations.html#t:outbox_name/0"},{"type":"type","title":"Carbonite.Migrations.outbox_option/0","doc":"","ref":"Carbonite.Migrations.html#t:outbox_option/0"},{"type":"type","title":"Carbonite.Migrations.patch/0","doc":"","ref":"Carbonite.Migrations.html#t:patch/0"},{"type":"type","title":"Carbonite.Migrations.prefix/0","doc":"","ref":"Carbonite.Migrations.html#t:prefix/0"},{"type":"type","title":"Carbonite.Migrations.table_name/0","doc":"","ref":"Carbonite.Migrations.html#t:table_name/0"},{"type":"type","title":"Carbonite.Migrations.trigger_config_key/0","doc":"","ref":"Carbonite.Migrations.html#t:trigger_config_key/0"},{"type":"type","title":"Carbonite.Migrations.trigger_option/0","doc":"","ref":"Carbonite.Migrations.html#t:trigger_option/0"},{"type":"type","title":"Carbonite.Migrations.up_option/0","doc":"","ref":"Carbonite.Migrations.html#t:up_option/0"},{"type":"module","title":"Carbonite.Multi","doc":"This module provides functions for dealing with audit trails in the context of Ecto.Multi.","ref":"Carbonite.Multi.html"},{"type":"function","title":"Carbonite.Multi.fetch_changes/2","doc":"Adds a operation to an `Ecto.Multi` to fetch the changes of the current transaction.\n\nUseful for returning all transaction changes to the caller.\n\nMulti step is called `:carbonite_changes`.\n\nSee `Carbonite.fetch_changes/2` for options.","ref":"Carbonite.Multi.html#fetch_changes/2"},{"type":"function","title":"Example - Carbonite.Multi.fetch_changes/2","doc":"Ecto.Multi.new()\n    |> Carbonite.Multi.insert_transaction(%{meta: %{type: \"create_rabbit\"}})\n    |> Ecto.Multi.insert(:rabbit, fn _ -> Rabbit.changeset(%{}) end)\n    |> Carbonite.Multi.fetch_changes()","ref":"Carbonite.Multi.html#fetch_changes/2-example"},{"type":"function","title":"Carbonite.Multi.insert_transaction/3","doc":"Adds an insert operation for a `Carbonite.Transaction` to an `Ecto.Multi`.\n\nMulti step is called `:carbonite_transaction`.\n\nSee `Carbonite.insert_transaction/3` for options.","ref":"Carbonite.Multi.html#insert_transaction/3"},{"type":"function","title":"Example - Carbonite.Multi.insert_transaction/3","doc":"Ecto.Multi.new()\n    |> Carbonite.Multi.insert_transaction(%{meta: %{type: \"create_rabbit\"}})\n    |> Ecto.Multi.insert(:rabbit, fn _ -> Rabbit.changeset(%{}) end)","ref":"Carbonite.Multi.html#insert_transaction/3-example"},{"type":"function","title":"Carbonite.Multi.override_mode/2","doc":"Sets the current transaction to \"override mode\" for all tables in the audit log.\n\nSee `Carbonite.override_mode/2` for options.","ref":"Carbonite.Multi.html#override_mode/2"},{"type":"type","title":"Carbonite.Multi.params/0","doc":"","ref":"Carbonite.Multi.html#t:params/0"},{"type":"type","title":"Carbonite.Multi.prefix/0","doc":"","ref":"Carbonite.Multi.html#t:prefix/0"},{"type":"type","title":"Carbonite.Multi.prefix_option/0","doc":"","ref":"Carbonite.Multi.html#t:prefix_option/0"},{"type":"module","title":"Carbonite.Outbox","doc":"A `Carbonite.Outbox` stores metadata for outboxes like the last processed transaction.\n\nThe `last_transaction_id` field defaults to zero, indicating that nothing has been\nprocessed yet.","ref":"Carbonite.Outbox.html"},{"type":"function","title":"Carbonite.Outbox.changeset/2","doc":"Builds an update changeset.","ref":"Carbonite.Outbox.html#changeset/2"},{"type":"type","title":"Carbonite.Outbox.memo/0","doc":"","ref":"Carbonite.Outbox.html#t:memo/0"},{"type":"type","title":"Carbonite.Outbox.name/0","doc":"","ref":"Carbonite.Outbox.html#t:name/0"},{"type":"type","title":"Carbonite.Outbox.t/0","doc":"","ref":"Carbonite.Outbox.html#t:t/0"},{"type":"module","title":"Carbonite.Query","doc":"This module provides query functions for retrieving audit trails from the database.","ref":"Carbonite.Query.html"},{"type":"function","title":"Carbonite.Query.changes/2","doc":"Returns an `t:Ecto.Query.t/0` that can be used to select changes for a single record.\n\nGiven an `t:Ecto.Schema.t/0` struct, this function builds a query that fetches all changes\nrecorded for it from the database, ordered ascending by their ID (i.e., roughly by\ninsertion date descending).","ref":"Carbonite.Query.html#changes/2"},{"type":"function","title":"Example - Carbonite.Query.changes/2","doc":"%MyApp.Rabbit{id: 1}\n    |> Carbonite.Query.changes()\n    |> MyApp.Repo.all()","ref":"Carbonite.Query.html#changes/2-example"},{"type":"function","title":"Options - Carbonite.Query.changes/2","doc":"* `carbonite_prefix` defines the audit trail's schema, defaults to `\"carbonite_default\"`\n* `table_prefix` allows to override the table prefix, defaults to schema prefix of the record\n* `preload` can be used to preload the transaction","ref":"Carbonite.Query.html#changes/2-options"},{"type":"function","title":"Carbonite.Query.current_transaction/1","doc":"Returns an `t:Ecto.Query.t/0` that can be used to select or delete the \"current\" transaction.\n\nThis function is useful when your tests run in a database transaction using Ecto's SQL sandbox.","ref":"Carbonite.Query.html#current_transaction/1"},{"type":"function","title":"Example: Asserting on the current transaction - Carbonite.Query.current_transaction/1","doc":"When you insert your `Carbonite.Transaction` record somewhere inside your domain logic, you do\nnot wish to return it to the caller only to be able to assert on its attributes in tests. This\nexample shows how you could assert on the metadata inserted.\n\n    # Test running inside Ecto's SQL sandbox.\n    test \"my test\" do\n      some_operation_with_a_transaction()\n\n      assert current_transaction_meta() == %{\"type\" => \"some_operation\"}\n    end\n\n    defp current_transaction_meta do\n      Carbonite.Query.current_transaction()\n      |> MyApp.Repo.one!()\n      |> Map.fetch(:meta)\n    end","ref":"Carbonite.Query.html#current_transaction/1-example-asserting-on-the-current-transaction"},{"type":"function","title":"Options - Carbonite.Query.current_transaction/1","doc":"* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`\n* `preload` - can be used to preload the changes, defaults to `false`","ref":"Carbonite.Query.html#current_transaction/1-options"},{"type":"function","title":"Carbonite.Query.outbox/2","doc":"Returns an `t:Ecto.Query.t/0` that selects a outbox by name.","ref":"Carbonite.Query.html#outbox/2"},{"type":"function","title":"Options - Carbonite.Query.outbox/2","doc":"* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.Query.html#outbox/2-options"},{"type":"function","title":"Carbonite.Query.outbox_done/1","doc":"Returns an `t:Ecto.Query.t/0` that selects all completely processed transactions.\n\n* If no outbox exists, this query returns all transactions.\n* If one or more outboxes exist, this query returns all transactions with an ID less than the\n  minimum of the `last_transaction_id` attributes of the outboxes.\n* Transactions are not ordered.","ref":"Carbonite.Query.html#outbox_done/1"},{"type":"function","title":"Options - Carbonite.Query.outbox_done/1","doc":"* `min_age` - the minimum age of a record, defaults to 300 seconds (set nil to disable)\n* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`","ref":"Carbonite.Query.html#outbox_done/1-options"},{"type":"function","title":"Carbonite.Query.outbox_queue/2","doc":"Returns an `t:Ecto.Query.t/0` that selects the next batch of transactions for an outbox.\n\n* Transactions are ordered by their ID ascending, so *roughly* in order of insertion.","ref":"Carbonite.Query.html#outbox_queue/2"},{"type":"function","title":"Options - Carbonite.Query.outbox_queue/2","doc":"* `min_age` - the minimum age of a record, defaults to 300 seconds (set nil to disable)\n* `limit` - limits the query in size, defaults to 100 (set nil to disable)\n* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`\n* `preload` - can be used to preload the changes, defaults to `true`","ref":"Carbonite.Query.html#outbox_queue/2-options"},{"type":"function","title":"Carbonite.Query.transactions/1","doc":"Returns an `t:Ecto.Query.t/0` that can be used to select transactions from the database.","ref":"Carbonite.Query.html#transactions/1"},{"type":"function","title":"Examples - Carbonite.Query.transactions/1","doc":"Carbonite.Query.transactions()\n    |> MyApp.Repo.all()\n\n    # Preload changes\n    Carbonite.Query.transactions(preload: true)\n    |> MyApp.Repo.all()","ref":"Carbonite.Query.html#transactions/1-examples"},{"type":"function","title":"Options - Carbonite.Query.transactions/1","doc":"* `carbonite_prefix` - defines the audit trail's schema, defaults to `\"carbonite_default\"`\n* `preload` - can be used to preload the changes, defaults to `false`","ref":"Carbonite.Query.html#transactions/1-options"},{"type":"type","title":"Carbonite.Query.changes_option/0","doc":"","ref":"Carbonite.Query.html#t:changes_option/0"},{"type":"type","title":"Carbonite.Query.current_transaction_option/0","doc":"","ref":"Carbonite.Query.html#t:current_transaction_option/0"},{"type":"type","title":"Carbonite.Query.disabled/0","doc":"","ref":"Carbonite.Query.html#t:disabled/0"},{"type":"type","title":"Carbonite.Query.outbox_done_option/0","doc":"","ref":"Carbonite.Query.html#t:outbox_done_option/0"},{"type":"type","title":"Carbonite.Query.outbox_queue_option/0","doc":"","ref":"Carbonite.Query.html#t:outbox_queue_option/0"},{"type":"type","title":"Carbonite.Query.prefix/0","doc":"","ref":"Carbonite.Query.html#t:prefix/0"},{"type":"type","title":"Carbonite.Query.prefix_option/0","doc":"","ref":"Carbonite.Query.html#t:prefix_option/0"},{"type":"type","title":"Carbonite.Query.preload_option/0","doc":"","ref":"Carbonite.Query.html#t:preload_option/0"},{"type":"type","title":"Carbonite.Query.transactions_option/0","doc":"","ref":"Carbonite.Query.html#t:transactions_option/0"},{"type":"module","title":"Carbonite.Transaction","doc":"A `Carbonite.Transaction` is the binding link between change records of tables.\n\nAs such, it contains a set of optional metadata that describes the transaction.","ref":"Carbonite.Transaction.html"},{"type":"function","title":"Carbonite.Transaction.changeset/1","doc":"Builds a changeset for a new `Carbonite.Transaction`.\n\nThe `:meta` map from the params will be merged with the metadata currently stored in the\nprocess dictionary.","ref":"Carbonite.Transaction.html#changeset/1"},{"type":"function","title":"Carbonite.Transaction.current_meta/0","doc":"Returns the currently stored metadata.","ref":"Carbonite.Transaction.html#current_meta/0"},{"type":"function","title":"Carbonite.Transaction.put_meta/2","doc":"Stores a piece of metadata in the process dictionary.\n\nThis can be useful in situations where you want to record a value at a system boundary (say,\nthe user's `account_id`) without having to pass it through to the database transaction.\n\nReturns the currently stored metadata.","ref":"Carbonite.Transaction.html#put_meta/2"},{"type":"type","title":"Carbonite.Transaction.id/0","doc":"","ref":"Carbonite.Transaction.html#t:id/0"},{"type":"type","title":"Carbonite.Transaction.meta/0","doc":"","ref":"Carbonite.Transaction.html#t:meta/0"},{"type":"type","title":"Carbonite.Transaction.t/0","doc":"","ref":"Carbonite.Transaction.html#t:t/0"},{"type":"module","title":"Carbonite.Trigger","doc":"A `Carbonite.Trigger` stores per table configuration for the change capture trigger.","ref":"Carbonite.Trigger.html"},{"type":"type","title":"Carbonite.Trigger.id/0","doc":"","ref":"Carbonite.Trigger.html#t:id/0"},{"type":"type","title":"Carbonite.Trigger.mode/0","doc":"","ref":"Carbonite.Trigger.html#t:mode/0"},{"type":"type","title":"Carbonite.Trigger.t/0","doc":"","ref":"Carbonite.Trigger.html#t:t/0"},{"type":"task","title":"mix carbonite.gen.initial_migration","doc":"Generates a sample migration that runs all migrations currently contained within Carbonite.","ref":"Mix.Tasks.Carbonite.Gen.InitialMigration.html"},{"type":"extras","title":"Changelog","doc":"## [0.14.0] - 2024-07-25\n\n**New migration patches:** 10\n\n#","ref":"changelog.html"},{"type":"extras","title":"Fixed - Changelog","doc":"- Improve performance of test setups by using a transaction-local variable in `override_mode/2` . The previous implementation could cause parallel tests to wait for row locks on the `triggers` table.\n- Replaced calls to `Ecto.Migration.timestamps/1` in migrations with explicit `add(:inserted_at, ...)` calls. This ensures that we don't be affected by users' `@migration_timestamps` options.\n\n## [0.13.0] - 2024-06-27\n\n**New migration patches:** 9\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fix `set_transaction_id` trigger not using the PK index on `transactions` due to being dependent on `CURRVAL` for every row. Replaced with a CTE the `EXIST` query is now an index-only scan as it was intended to be. As a consequence, scalability of inserts into the `transactions` table has been greatly improved.\n\n## [0.12.1] - 2024-05-21\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fix incomplete rollback logic in migration 8, leading to leftover procedures in the carbonite prefix when rolling back all the way (e.g. `Carbonite.Migrations.down(8..1)`). (#100, @xadhoom)\n\n## [0.12.0] - 2024-05-17\n\n**New migration patches:** 8\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Make the trigger function correctly mask the `filtered_columns` in the `changed_from` JSON object if `store_changed_from` is set. (#96)\n\n⚠️ Unfortunately, this bug caused Carbonite to persist potentially sensitive data in the `changed_from` object even though they were listed in the `filtered_columns`. All users who combine `store_changed_from: true` with a `filtered_columns` setting should upgrade and assess the impact of this potential data leak.\n\n## [0.11.1] - 2024-04-25\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Migration: Creating table `changes` failed when Ecto's `@migration_foreign_key` is set to `[column:  ]`.\n\n## [0.11.0] - 2023-11-14\n\n**New migration patches:** 7\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- Add a new index on `transaction_id` in the `changes` table, to speed up queries for a transaction's changes and the `purge` operation. The existing `changes_transaction_id_index` (on `transaction_xact_id` column) has been renamed to `changes_transaction_xact_id_index`.\n- Allow ranges of migration patches in `Carbonite.Migrations.up/2` / `down/2`.\n\n## [0.10.0] - 2023-09-20\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fixed a minor incompatibility with recently released Postgres 16.0.\n\n## [0.9.0] - 2023-04-27\n\n**New migration patches:** 6\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Correctly detect changes to array fields.\n\nPreviously, detection of changes was done via the `@>` operator to test \"containment\" of a `{col: old_value}` JSON object in the JSON object of the new record. Unfortunately, Postgres' \"jsonb containment\" (see [docs](https://www.postgresql.org/docs/15/datatype-json.html#JSON-CONTAINMENT)) views array subsets as contained within their subsets, as well as arrays in different orders to be contained within each other. Both of these cases we want to track as a changed value.\n\n⚠️ This bug caused Carbonite to not identify a changed array field correctly, meaning it may not have been listed in the `changed` and `changed_from` columns of the `Carbonite.Change` record. Unfortunately, this also means that **versions may have been discarded entirely** if no other fields of a record were updated at the same time.\n\n## [0.8.0] - 2023-03-27\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Add `:to` option to `Carbonite.override_mode/2` to specify an explicit target mode. Useful for toggling the mode around test fixtures.\n\n## [0.7.2] - 2023-03-07\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fix `Carbonite.process/4` success typing by explicitly picking options.\n\n## [0.7.1] - 2023-02-15\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fix `carbonite_prefix` on `Query.outbox_done/2`. Apply the prefix on the nested outbox query as well.\n\n## [0.7.0] - 2023-02-15\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"* `Carbonite.override_mode/2` and most functions in `Carbonite.Query` were broken when using a non-default `carbonite_prefix` option. Fixed by moving the `prefix` option onto the `from` expression.\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Added `:initially` option to `create_trigger/2` to create triggers with `IMMEDIATE DEFERRED` constraint option. This allows to conditionally insert the `Carbonite.Transaction` record at the end of the transaction. In order to use this for already existing triggers, you need to drop them (`drop_trigger/2`) and re-create them.\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"* Made the changes trigger `DEFERRABLE`. As part of the `:initially` option of `create_trigger/2`, we chose to make triggers `DEFERRABLE` by default. Again, for any existing triggers, this won't take effect, but newly created triggers will use this constraint option.\n* Dropped support for `preload: atom | [atom]`-style of specifying preloads on a query in `Carbonite.Query`.\n\n## [0.6.0] - 2022-10-12\n\n**New migration patches:** 5\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"* Optional tracking of previous data in changes. Set the `store_changed_from` trigger option.\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"* Added `@schema_prefix \"carbonite_default\"` on all schemas. This will enable the manual usage of `Repo.insert/2` & friends on transactions without specifying a prefix, when using the default carbonite prefix.\n\n## [0.5.0] - 2022-02-25\n\n**New migration patches:** 4\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Switch to normal identity column on `transactions` - Changelog","doc":"* The `id` column on `transactions` has been replaced with an ordinary autoincrementing integer PK, filled from a sequence. Next to it a new `xact_id` column continues to store the transaction id (from `pg_current_xact_id`). Both values used together ensure that, first the `id` is monotonically increasing and survives a backup restore (see issue #45), and second the `changes` records can still only be inserted within the same transaction.\n\n#","ref":"changelog.html#switch-to-normal-identity-column-on-transactions"},{"type":"extras","title":"Added - Changelog","doc":"* `Carbonite.Migrations.insert_migration_transaction/1` and its macro friend, `insert_migration_transaction_after_begin`, help with data migrations.\n* `Carbonite.fetch_changes` returns the changes of the current (ongoing) transaction.\n\n## [0.4.0] - 2021-11-07\n\n**New migration patches:** 2, 3\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Switch to top-level API with `repo` param - Changelog","doc":"* `Carbonite.override_mode/2` (kept a wrapper in `Carbonite.Multi`)\n* `Carbonite.insert_transaction/3` (kept a wrapper in `Carbonite.Multi`)\n* `Carbonite.process/4` (previously `Carbonite.Outbox.process/3` with major changes)\n* `Carbonite.purge/2` (previously `Carbonite.Outbox.purge/1` with major changes)\n\n#","ref":"changelog.html#switch-to-top-level-api-with-repo-param"},{"type":"extras","title":"Big outbox overhaul - Changelog","doc":"* Split into query / processing\n* Simplify processing\n* No more transaction\n* New capabilities: memo, halting, chunking\n\n#","ref":"changelog.html#big-outbox-overhaul"},{"type":"extras","title":"Migration versioning - Changelog","doc":"* Explicit for now with `Carbonite.Migrations.up(non_neg_integer())`\n* `Carbonite.Migrations.install_schema/1` is now `Carbonite.Migrations.up/2`\n* `Carbonite.Migrations.put_trigger_option/4` to ensure old migrations continue to work\n* At the same time removed long configuration statement from `Carbonite.Migrations.install_trigger/1`, so this does not need to be versioned and continues to work\n* Mix task for generating the \"initial\" migration\n\n#","ref":"changelog.html#migration-versioning"},{"type":"extras","title":"Other Changes - Changelog","doc":"* Optionally derive Jason.Encoder for `Carbonite.Transaction` and `Carbonite.Change`\n* Made all prefix options binary-only (no atom) as `Ecto.Query.put_query_prefix/2` only accepts strings\n\n## [0.3.1] - 2021-10-23\n\n#","ref":"changelog.html#other-changes"},{"type":"extras","title":"Added - Changelog","doc":"* `table_prefix` option to `Query.changes/2` allows to override schema prefix of given record\n* `Query.transactions/1` query selects all transactions\n\n## [0.3.0] - 2021-10-22\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Added - Changelog","doc":"* `Carbonite.Migrations.drop_tables/1` allows to drop the carbonite audit trail without removing the schema\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"* Renamed the option that can be passed to `Carbonite.Migrations.drop_schema/1` from `prefix` to `carbonite_prefix`\n* Changed `Carbonite.Migrations.drop_schema/1` to also drop the tables\n* Made `Carbonite.Multi.insert_transaction/3` ignore conflicting `INSERT`s within the same transaction\n* Also, changed `Carbonite.Multi.insert_transaction/3` to always reloads all fields from the database after insertion, immediately returning the JSONinified `meta` payload\n\n#","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fixed ignore mode when `override_transaction_id` is NULL\n\n## [0.2.1] - 2021-10-10\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Fixed - Changelog","doc":"* Fixed broken documentation\n\n## [0.2.0] - 2021-10-10\n\n#","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"* Store primary key(s) on changes table and index them\n* Add `Carbonite.Query` module\n  - `current_transaction/2` allows to fetch the ongoing transaction (for sandbox tests)\n  - `changes/2` allows to fetch the changes of an individual source record\n* Update Postgrex to 0.15.11 and drop local `Xid8` type\n* Add `mode` field to trigger (capture or ignore)\n* Add \"override mode\" reversing the `mode` option for the current transaction to enable/disable capturing on demand (e.g. in tests)\n* Add filtered columns\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"* Moved top-level functions to nested modules `Transaction` and `Multi`\n* Made `table_pk` be `NULL` when `primary_key_columns` is an empty array\n* Default `primary_key_columns` to `[\"id\"]`\n* Renamed `prefix` option to `carbonite_prefix` on `install_schema/2` for consistency\n\n## [0.1.0] - 2021-09-01\n\n* Initial release.","ref":"changelog.html#changed"},{"type":"extras","title":"License","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n   APPENDIX: How to apply the Apache License to your work.\n\n      To apply the Apache License to your work, attach the following\n      boilerplate notice, with the fields enclosed by brackets \"[]\"\n      replaced with your own identifying information. (Don't include\n      the brackets!)  The text should be enclosed in the appropriate\n      comment syntax for the file format. We also recommend that a\n      file or class name and description of purpose be included on the\n      same \"printed page\" as the copyright notice for easier\n      identification within third-party archives.\n\n   Copyright 2021-2022 Bitcrowd GmbH\n\n   Licensed under the Apache License, Version 2.0 (the \"License\");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an \"AS IS\" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.","ref":"license.html"}],"content_type":"text/markdown"}